#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include "GlobFactory.h"
#include "RandomStream.h"
#include "RelativePoint.h"
#include "SphereSurface.h"
#include "pi.h"
#include "nsew.h"


GlobFactory::GlobFactory( void ) {
    basisMatrix_ = catmull_rom;
    shader_ = new SurfaceShader();
}

GlobFactory::GlobFactory( Shader* shader ) {
    basisMatrix_ = catmull_rom;
    shader_ = shader;
}

// Define a new surface relative to some
// existing surface.

 PatchSurface*
GlobFactory::glob(
    Surface* core,
    int inum,
    int jnum,
    double thickness,
    Vary* vary,
    char* name
) {
    PatchSurface* glob = new PatchSurface( name, shader_ );
    RelativePoint* point[256][256+1];
    CubicPatch* patch[256][256];
    double s[256];
    double t[256+3];

    assert( inum <= 256 );
    assert( jnum <= 256 );

    // For now at least, inum must be even:
    assert( (inum & 0) == 0 );

    // Precompute all parametric coords we'll
    // need for the sphere.  For inum==jnum==4:
    //   'i' goes around the equator in four steps.
    //   'j' needs to have values for:
    //       North and south pole
    //       Equator
    //       Midway between each pole and equator
    //       Two redundant wrap-around convenience points.
    //       Total: 7 points spaced PI/4 apart:
    int      i,j;
    for     (i = 0;   i < inum  ;   ++i) {
	for (j = 0;   j < jnum+3;   ++j) {

            // Remember parametric coords in [0,1) for above:
            s[i]  = (1.0 / (double)inum) * (double)i;
            t[j]  = (1.0 / (double)jnum) * (double)(j-1);

	}
    }

    // Create all the control points.
    // There is one per patch, in general,
    // except that the north and south poles
    // have one point only instead of a row:
    for     (i = 0;  i < inum; ++i) {
        for (j = 1;  j < jnum; ++j) {
	    double s = (double)i/(double)inum;
	    double t = (double)j/(double)jnum;
	    assert(i<256 && j <=256);
            point[i][j] = new RelativePoint( core, vary->rst( Rst( thickness, s, t ), core ) );
            glob->addPoint( point[i][j] );

        }
    }
    {   RelativePoint* southPole =  new RelativePoint( core, vary->rst( Rst( thickness, 0.0, 0.0 ), core ) );
        glob->addPoint( southPole );
        for (i = 0;  i < inum  ; ++i) {
	    assert(i<256);
            point[i][0] = southPole;
        }
    }
    {   RelativePoint* northPole =  new RelativePoint( core, vary->rst( Rst( thickness, 0.0, 1.0 ), core ) );
        glob->addPoint( northPole );
        for (i = 0;  i < inum  ; ++i) {
	    assert(i<256 && jnum <=256 );
            point[i][jnum] = northPole;
        }
    }


    // Over all patches to be created (a inum x jnum grid):
    for     (i = 0;  i < inum; ++i) {  // i 0==backright 1==frontright 2==frontleft 3==backleft
        for (j = 0;  j < jnum; ++j) {  // j==0 is bottom cap, j==3 is top

            // Create the patch:
            St minSt( (i+0) * (1.0/(double)inum), (j+0) * (1.0/(double)jnum) );
            St maxSt( (i+1) * (1.0/(double)inum), (j+1) * (1.0/(double)jnum) );
            patch[i][j] = new CubicPatch(); 
            patch[i][j]->basisMatrix( basisMatrix_ );
            patch[i][j]->parity( (i+j) & 1 );
            patch[i][j]->minSt( minSt );
            patch[i][j]->maxSt( maxSt );
            glob->insert( patch[i][j] );

            // Name it:
            char buf[ 132 ];
            sprintf( buf, "patch%d.%d", i, j );
            patch[i][j]->name( buf );

            // Point it to the 16 control
            // points it needs:
	    int m,n;
	    for     (m = 0; m < 4; ++m) { 
		for (n = 0; n < 4; ++n) { 
		    int    u = (i+inum+m-1)%inum;
		    int    v =  j+n-1;
                    if (v ==     -1) { v =      1; u = (u+inum/2) % inum; }
                    if (v == jnum+1) { v = jnum-1; u = (u+inum/2) % inum; }
                    assert(u >= 0 && u<256 && v >= 0 && v <=256 );
                    patch[i][j]->controlPoint( m, n, point[u][v] );
		}
	    }
	}
    }

    // Now that all patches are created, tell each about its neighbors:
    for     (i = 0;  i < inum; ++i) {  // i 0==backright 1==frontright 2==frontleft 3==backleft
        for (j = 0;  j < jnum; ++j) {  // j==0 is bottom cap, j==3 is top

	    patch[i][j]->neighbor( EAST, patch[(i+      1 )%inum][j] );
	    patch[i][j]->neighbor( WEST, patch[(i+(inum-1))%inum][j] );

	    if (j)    patch[i][j]->neighbor(         SOUTH, patch[i][j-1] );
            else      patch[i][j]->neighbor( TWISTED_SOUTH, patch[(i + inum/2)  % inum][j] );

	    if (j+1<jnum) patch[i][j]->neighbor(         NORTH, patch[i][j+1] );
            else          patch[i][j]->neighbor( TWISTED_NORTH, patch[(i+ inum/2)   % inum][j] );
	}
    }

    glob->checkNeighbors();

    return glob;
}

 BasisMatrix*
GlobFactory::basisMatrix( void ) {
   return basisMatrix_;
}

 void
GlobFactory::basisMatrix( BasisMatrix* basisMatrix ) {
   basisMatrix_ = basisMatrix;
}

 Shader*
GlobFactory::shader( void ) {
   return shader_;
}

 void
GlobFactory::shader( Shader* shader ) {
   shader_ = shader;
}

// Not currently used:

//  static bool
// approx( double a, double b ) {
//     if (a < 0.00001 && a > -0.00001) a = 0.0;
//     if (b < 0.00001 && b > -0.00001) b = 0.0;
//     if (a > b) { double t = a; a = b; b = t; }
//     if (b == 0.0)  return a == 0.0;
//     return a/b > 0.999;
// }
// 
//  static bool
// matchRst( Rst a, Rst b ) {
//     // Remember that if t==0 or t==1, then
//     // we have a pole point and s is irrelevant:
//     return  approx( a.r(), b.r() )
//         &&  (a.t() == 0.0 || a.t() == 1.0 || approx( a.s(), b.s() ))
//         &&  approx( a.t(), b.t() )
//     ;
// }
