#ifndef XYZW_H
#define XYZW_H

/* This is our main type for passing actual */
/* point values around.  We always pass it  */
/* by value, not by reference:              */

#include <stdio.h>

class Xyzw {

  public:
    explicit Xyzw(): x_(0.0), y_(0.0), z_(0.0), w_(1.0) {}
    explicit Xyzw( double x, double y ): x_(x), y_(y), z_(0.0), w_(1.0) {}
    explicit Xyzw( double x, double y, double z ): x_(x), y_(y), z_(z), w_(1.0) {}
    explicit Xyzw( double x, double y, double z, double w ): x_(x), y_(y), z_(z), w_(w) {
        if (w_ != 0.0 && w_ != 1.0) {
            double inv = 1.0 / w_;
            x_ *= inv;
            y_ *= inv;
            z_ *= inv;
            w_  = 1.0;
        }
    }

    inline const Xyzw operator +( Xyzw const & ) const;
    inline const Xyzw operator -( Xyzw const & ) const;
    inline double operator *( Xyzw const & ) const; // Dot prod
    inline const Xyzw operator *( double const ) const; // Scale
    inline const Xyzw operator ^( Xyzw const & ) const; // Cross prod
    inline bool operator ==( Xyzw const & ) const; // Equality

    inline const Xyzw xNeg( void );
    inline const Xyzw yNeg( void );
    inline const Xyzw zNeg( void );

    inline double x() const;
    inline double y() const;
    inline double z() const;
    inline double w() const;

    double length(void) const;
    Xyzw const normalize(void) const;

    void print(void) const;
    void println(void) const;

  private:
    Xyzw const standardize(void) const;

    double x_;
    double y_;
    double z_;
    double w_;
};


 inline const Xyzw
Xyzw::standardize( void ) const return ret {
    ret.x_ = x_;
    ret.y_ = y_;
    ret.z_ = z_;
    ret.w_ = w_;
    if (w_ != 0.0 && w_ != 1.0) {
        double inv = 1.0 / w_;
        ret.x_ *= inv;
        ret.y_ *= inv;
        ret.z_ *= inv;
        ret.w_  = 1.0;
    }
    return ret;
}

 inline const Xyzw
Xyzw::operator +( Xyzw const & o ) const {
    Xyzw ret(
        x_ + o.x_,
        y_ + o.y_,
        z_ + o.z_,
        w_ * o.w_
    );
    return ret.standardize();
}

 inline const Xyzw
Xyzw::operator -( Xyzw const & o ) const {
    Xyzw ret(
        x_ - o.x_,
        y_ - o.y_,
        z_ - o.z_,
        w_ * o.w_
    );
    return ret.standardize();
}

 // Dot product:
 inline double
Xyzw::operator *( Xyzw const & o ) const {
    double xyz = (
        x_ * o.x_ +
        y_ * o.y_ +
        z_ * o.z_
    );
    double w = w_ * o.w_;
    if (w != 0.0) xyz /= w;
    return xyz;
}

 // Scale:
 inline const Xyzw
Xyzw::operator *( double const k ) const {
    Xyzw ret(
        x_ * k,
        y_ * k,
        z_ * k,
        w_
    );
    return ret;
}

 // Cross product:
 inline const Xyzw
Xyzw::operator ^( Xyzw const & o ) const {
    Xyzw ret(
        y_ * o.z_ - z_ * o.y_,
        z_ * o.x_ - x_ * o.z_,
        x_ * o.y_ - y_ * o.x_,
        w_ * o.w_
    );
    return ret.standardize();
}

 inline bool
Xyzw::operator ==( Xyzw const & o ) const {
    return 
        x_ == o.x_ &&
        y_ == o.y_ &&
        z_ == o.z_ &&
        w_ == o.w_;
}

inline double Xyzw::x(void) const { return x_; }
inline double Xyzw::y(void) const { return y_; }
inline double Xyzw::z(void) const { return z_; }
inline double Xyzw::w(void) const { return w_; }

inline const Xyzw Xyzw::xNeg(void) { return Xyzw( -x_,  y_,  z_,  w_ ); }
inline const Xyzw Xyzw::yNeg(void) { return Xyzw(  x_, -y_,  z_,  w_ ); }
inline const Xyzw Xyzw::zNeg(void) { return Xyzw(  x_,  y_, -z_,  w_ ); }

#endif /* XYZW_H */
