#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include "AbsolutePoint.h"
#include "Animation.h"
#include "GlobFactory.h"
#include "BasisMatrix.h"
#include "CubicPatch.h"
#include "CubicSpline.h"
#include "Hash.h"
#include "PatchSurface.h"
#include "RelativePoint.h"
#include "Rgba.h"
#include "SphereSurface.h"
#include "SurfaceShader.h"
#include "VaryRandomlyAlongNormal.h"
#include "VaryOnSOnly.h"
#include "VaryOnTOnly.h"
#include "VaryWithSymmetry.h"
#include "XMirrorPoint.h"
#include "Rst.h"
#include "St.h"
#include "Xyzw.h"
#include "glob.h"

static bool symmetry = 0;
static bool varyOnT  = 0;

static bool mirrorS  = 0;
static bool mirrorT  = 0;
static bool mirrorX  = 0;
static bool mirrorY  = 0;
static bool mirrorZ  = 0;

static bool ignoreS  = 1;
static bool ignoreT  = 1;
static bool ignoreX  = 1;
static bool ignoreY  = 1;
static bool ignoreZ  = 1;

double lightradius = 0.0;
int caustics = 0;

Rgba rgba = Rgba( 1.0, 1.0, 1.0 );

static char* save_file = 0;
static FILE* save_fd = 0;

static char* load_file = 0;
static FILE* load_fd = 0;

char tif_file[ 132 ];

static char* rib_file = 0;
static FILE* rib_fd = 0;

int nofloor = 0;
int nofill = 0;
int debug = 0;
int chequer_all_surfaces = 0;
static BasisMatrix* basisMatrix;

 void
init( void ) {
    Hash_init();
    BasisMatrix_init();
}

 static bool
approx(
    double x,
    double y
) {
    if (x > y) { double t = x; x = y; y = t; }
    return (x / y) > 0.95;
}

 void
test_Xyzw( void ) {
    Xyzw xyzw1 = Xyzw( 1.0, 0.0, 1.0, 1.0 );
    Xyzw xyzw2 = Xyzw( 0.0, 1.0, 0.0, 1.0 );
    Xyzw xyzw3 = xyzw1 + xyzw2;

    assert( approx( 1.0, xyzw3.x() ));
    assert( approx( 1.0, xyzw3.y() ));
    assert( approx( 1.0, xyzw3.z() ));
    assert( approx( 1.0, xyzw3.w() ));
}


 void
test_CubicSpline( void ) {
    CubicSpline s(
        basisMatrix,
	new AbsolutePoint( Xyzw( 0.0, 0.0, 0.0 ) ),
        new AbsolutePoint( Xyzw( 1.0, 1.5, 0.0 ) ),
        new AbsolutePoint( Xyzw( 2.0, 2.0, 0.0 ) ),
        new AbsolutePoint( Xyzw( 3.0, 3.0, 0.0 ) )
    );
    double d;
    for (d = 0.0; d < 1.001;  d += 0.1) {
        Xyzw dd = s.value( d );
        assert( approx( d+1.0, dd.x() ));
    }
}

 void
test_CubicPatch( void ) {
    CubicPatch p(
        basisMatrix,
	new AbsolutePoint(Xyzw( 0.0, 0.0 )),  new AbsolutePoint(Xyzw( 0.0, 1.0 )), new AbsolutePoint(Xyzw( 0.0, 2.0 )), new AbsolutePoint(Xyzw( 0.0, 3.0 )),
	new AbsolutePoint(Xyzw( 1.0, 0.0 )),  new AbsolutePoint(Xyzw( 1.0, 1.0 )), new AbsolutePoint(Xyzw( 1.0, 2.0 )), new AbsolutePoint(Xyzw( 1.0, 3.0 )),
	new AbsolutePoint(Xyzw( 2.0, 0.0 )),  new AbsolutePoint(Xyzw( 2.0, 1.0 )), new AbsolutePoint(Xyzw( 2.0, 2.0 )), new AbsolutePoint(Xyzw( 2.0, 3.0 )),
	new AbsolutePoint(Xyzw( 3.0, 0.0 )),  new AbsolutePoint(Xyzw( 3.0, 1.0 )), new AbsolutePoint(Xyzw( 3.0, 2.0 )), new AbsolutePoint(Xyzw( 3.0, 3.0 )),
        St( 0.0, 0.0 ),
        St( 1.0, 1.0 )
    );
    double s = 0.5;
    double t = 0.5;
    Xyzw dd = p.value( s, t );
    assert( approx( 1.5, dd.x()     ));
    assert( approx( 1.5, dd.y()     ));
    assert( approx( 1.0, dd.z()+1.0 ));
}

 Xyzw
test_RelativePoint( void ) {
    SphereSurface p( 1.0, 1.0, 1.0 );
    Rst rst( 0.0, 0.0, 1.0 );
    RelativePoint pp( &p, rst );
    Xyzw ee = pp.value();
    return ee;
}

 Xyzw
test_XMirrorPoint( void ) {
    AbsolutePoint fp( Xyzw( 1.0, 2.0, 3.0 ) );
    XMirrorPoint mp( &fp );
    Xyzw x = mp.value();
    return x;
}

static char* text = 0;

#define MAX_LAYERS 7
static unsigned int layers = 3;
struct surf_spec {
    int    inum;
    int    jnum;
    double thickness;
    double variation;
    char*  name;
    char*  seed;
} surf[ MAX_LAYERS ] = {
    {   4,   4, 0.0000, 1.0, "first",   "seed" },
    {   8,   8, 0.0000, 1.0, "second",  "seed" },
    {  16,  16, 0.0000, 1.0, "third",   "seed" },
    {  32,  32, 0.0000, 1.0, "fourth",  "seed" },
    {  64,  64, 0.0000, 1.0, "fifth",   "seed" },
    { 128, 128, 0.0000, 1.0, "sixth",   "seed" },
    { 256, 256, 0.0000, 1.0, "seventh", "seed" }
};

 void
maybe_open_save_fd( void ) {
    if (!save_file) return;
    save_fd = fopen(save_file,"w");
    if(!save_fd) {
        fprintf(stderr,"Couldn't create %s, skipping state save\n",save_file);
        return;
    }
}
 void
maybe_save_state( PatchSurface* glob ) {
    if (!save_file) return;
    glob->saveState( save_fd, "" );    
}
 void
maybe_close_save_fd( void ) {
    if (!save_file) return;
    fclose(save_fd);
}

 void
maybe_open_load_fd( void ) {
    if (!load_file) return;
    load_fd = fopen(load_file,"r");
    if(!load_fd) {
        fprintf(stderr,"Couldn't open %s, aborting run\n",load_file);
        assert(0);
    }
}
 void
maybe_load_state( PatchSurface* glob ) {
    if (!load_file) return;
    glob->loadState( load_fd, "" );    
}
 void
maybe_close_load_fd( void ) {
    if (!load_file) return;
    fclose(load_fd);
}

#define MEAN(i,j)  sqrt( (double) ((i) * (j)) )

 void
test_GlobFactory( void ) {
    Animation* animation = new Animation;
    Shader * shader = new SurfaceShader( text, rgba );
    GlobFactory * gf = new GlobFactory( shader );
    Surface* core = new SphereSurface( 1.0, 1.0, 1.0 );
    gf->basisMatrix( basisMatrix );
    PatchSurface* glob = 0;
    unsigned int i;
    maybe_open_save_fd();
    maybe_open_load_fd();
    for (i = 0;  i < layers;  ++i) {
        struct surf_spec s = surf[i];
        double scale = 4.0 * s.variation / MEAN( s.inum, s.jnum );

        Vary* vary;
        if (varyOnT) {
            vary = new VaryOnTOnly( scale, hash->to_0_1( s.seed ,s.name ) );
        } else if (symmetry) {
            vary = new VaryWithSymmetry(
                scale,
                hash->to_0_1( s.seed, s.name ),
                mirrorS, mirrorT, mirrorX, mirrorY, mirrorZ,
                ignoreS, ignoreT, ignoreX, ignoreY, ignoreZ
            );
        } else {
            vary = new VaryRandomlyAlongNormal( RandomStream( s.seed, s.name ), scale );
//          case SYMMETRY_AXIAL:  vary = new VaryOnSOnly( scale, hash->to_0_1( s.seed, s.name ) ); break;
//          case SYMMETRY_RADIAL: vary = new VaryOnTOnly( scale, hash->to_0_1( s.seed ,s.name ) ); break;
        }

        char namebuf[ 132 ];
        sprintf( namebuf, "Layer%d", i );
        glob = gf->glob(
	    i ? glob : core,
	    s.inum,
	    s.jnum,
	    4.0 * s.thickness / MEAN( s.inum, s.jnum ),
            vary,
            namebuf
        );
        maybe_save_state( glob );
        maybe_load_state( glob );
    }
    maybe_close_save_fd();
    maybe_close_load_fd();
    animation->insert( glob );
    animation->rib( rib_fd ? rib_fd : stdout );
}

 void
test( void ) {
    test_Xyzw();
    test_RelativePoint();
    test_CubicSpline();
    test_CubicPatch();
    test_XMirrorPoint();
    test_GlobFactory();
}

 static void
select_plastic(
    void
) {
    text = "Surface \"plastic\" \"Ka\" [1.0] \"Kd\" [0.5] \"Ks\" [0.5]  \"roughness\" [0.1]\n";
}

 static void
select_constant(
    void
) {
    text = "Surface \"constant\"\n"; 
}

 static void
select_roughmetal(
    void
) {
    text = "Surface \"roughmetal\" \"roughness\" [0.3] \"Ks\" [1.5]\n";
}

 static void
select_shinymetal(
    void
) {
    text = "Surface \"shinymetal\"\n";
}

 static void
select_shiny(
    void
) {
//  text = "Surface \"shiny\" \"roughness\" [0.3] \"Ks\" [1.5] \"Kd\" [0.15] \"Kr\" [0.4]\n";
    text = "Surface \"shiny\" \"roughness\" [0.3] \"Ks\" [1.5] \"Kd\" [0.15] \"Kr\" [0.4]\n";
}

 static void
select_shinyplastic(
    void
) {
    text = "Surface \"shinyplastic\" \"roughness\" [0.3] \"Ks\" [1.5] \"Kd\" [0.15] \"Kr\" [0.4]\n";
}

 static void
select_clay(
    void
) {
    text = "Surface \"clay\"\n";
}

 static void
select_matte(
    void
) {
    text = "Surface \"matte\"\n";
}

 static void
select_veinedmarble(
    void
) {
    text = "Surface \"veinedmarble\"\n";
}

 static void
select_greenmarble(
    void
) {
    text = "Surface \"greenmarble\"\n";
}

 static void
select_wood2(
    void
) {
    text = "Surface \"wood2\"\n";
}

 static void
select_oak(
    char* cmd
) {
    if (!strcmp(cmd, "roughoak")) {
       text = "Surface \"oak\" \"float divotdepth\" [0.007]\n";
    } else {
       text = "Surface \"oak\" \"float divotdepth\" [0] \"float roughness\" [0.075] \"float Ks\" [1.4] \"float Kd\" [0.3] \"float grainy\" [0.5] \"color Cdarkwood\" [0.07 0.04 0.014]\n";
    }
}

 static void
select_glass(
    void
) {
    text =
	"Declare \"transmitcolor\" \"color\"\n"
	"Declare \"blur\" \"float\"\n"
	"Declare \"refblur\" \"float\"\n"
	"Declare \"eta\" \"float\"\n"
	"Declare \"roughness\" \"float\"\n"
#ifdef CAUSTIC // Never got this to work
        "Attribute \"radiosity\" \"specularcolor\" [1 1 1]\n"
        "Attribute \"radiosity\" \"refractioncolor\" [1 1 1]\n"
        "Attribute \"radiosity\" \"refractionindex\" [2.2]\n"
#endif
	"Surface \"glass\""
//	   " \"transmitcolor\" [1.0 1.0 1.0]"
//	   " \"Ka\" [0.0]"
//	   " \"Kd\" [0.0]"
//	   " \"Ks\" [1.0]"
//	   " \"roughness\" 0.0"
//	   " \"blur\" 0.0"
//	   " \"refblur\" 0.0"
//	   " \"eta\" 2.2"
    "\n";
}

 static void
select_superplank(
    void
) {
   text =  "Declare \"Kr\" \"float\"\n"
          "Surface \"superplank\" \"Kr\" 1\n";
}

 static void
select_screen(
    void
) {
   text =
       "Declare \"frequency\" \"float\"\n"
       "Surface \"screen\" \"frequency\" 5\n";
}

 static void
select_color(
    char* text
) {
   float red = 1.0;
   float green = 1.0;
   float blue = 1.0;
   sscanf( text, "%g,%g,%g", &red, &green, &blue );
   rgba = Rgba( red, green, blue );
}



 static void
usage( void ) {
    fprintf(stderr, "usage: glob [options] outfile.rib\n" );
    fprintf(stderr, "  General options:\n" );
    fprintf(stderr, "    --color=0.3,0.4,0.5\n" );
    fprintf(stderr, "    --nofloor\n" );
    fprintf(stderr, "    --nofill\n" );
    fprintf(stderr, "    --layers=1-%d\n", MAX_LAYERS );
    fprintf(stderr, "    --seed=anyword\n" );
    fprintf(stderr, "    --seed1=anyword\n" );
    fprintf(stderr, "      ...\n" );
    fprintf(stderr, "    --seed%d=anyword\n", MAX_LAYERS );
    fprintf(stderr, "    --lumpy=0.0-2.0 # Overall lumpiness\n" );
    fprintf(stderr, "    --lumpy1=0.0-2.0 # Lumpiness for given layer\n" );
    fprintf(stderr, "      ...\n" );
    fprintf(stderr, "    --lumpy%d=0.0-2.0\n", MAX_LAYERS );
    fprintf(stderr, "    --symmetry=[x][y][z][X][Y][Z] # xyz=mirror XYZ=axial/radial\n" );
    fprintf(stderr, "  State save/restore options:\n" );
    fprintf(stderr, "    --savefile=foo.glob\n" );
    fprintf(stderr, "    --loadfile=foo.glob\n" );
    fprintf(stderr, "  Lighting options:\n" );
    fprintf(stderr, "    --lightradius=0.1-5000\n" );
    fprintf(stderr, "    --caustics\n" );
    fprintf(stderr, "  Cubic spline options:\n" );
    fprintf(stderr, "    --bspline\n" );
    fprintf(stderr, "  Debug options:\n" );
    fprintf(stderr, "    --checker\n" );
    fprintf(stderr, "    --debug\n" );
    fprintf(stderr, "  Shader options:\n" );
    fprintf(stderr, "    --plastic\n" );
    fprintf(stderr, "    --constant\n" );
    fprintf(stderr, "    --roughmetal\n" );
    fprintf(stderr, "    --shinymetal\n" );
    fprintf(stderr, "    --shiny\n" );
    fprintf(stderr, "    --shinyplastic\n" );
    fprintf(stderr, "    --clay\n" );
    fprintf(stderr, "    --matte\n" );
    fprintf(stderr, "    --veinedmarble\n" );
    fprintf(stderr, "    --greenmarble\n" );
    fprintf(stderr, "    --wood2\n" );
    fprintf(stderr, "    --oak\n" );
    fprintf(stderr, "    --roughoak\n" );
    fprintf(stderr, "    --glass\n" );
    fprintf(stderr, "    --superplank\n" );
    fprintf(stderr, "    --screen\n" );
    fprintf(stderr, "Typical usage:\n");
    fprintf(stderr, "   ./glob --greenmarble test.rib; rendrib test.rib ; iv test.tif\n");
    exit(1);
}

 static void
do_lumpy( const char* txt ) {
    double lumpy = atof( txt );
    assert( lumpy >= 0.0 && lumpy < 10.0 );
    int i;
    for (i = 0; i < MAX_LAYERS; ++i)  surf[i].variation = lumpy;
}

 static void
do_lumpyn( const char* txt, int n ) {
    double lumpy = atof( txt );
    assert( lumpy >= 0.0 && lumpy < 10.0 );
    surf[n].variation = lumpy;
}

 static void
do_thickness( const char* txt ) {
    double thickness = atof( txt );
    assert( thickness >= -1.0 && thickness < 10.0 );
    int i;
    for (i = 0; i < MAX_LAYERS; ++i)  surf[i].thickness = thickness;
}

 static void
do_lightradius( const char* txt ) {
    double radius = atof( txt );
    assert( radius > 0.01 && radius < 100000.0 );
    lightradius = radius;
}

 static void
do_symmetry( const char* txt ) {
    char c;
    symmetry = 1;
    while (c = *txt++) {
        switch (c) {
        case 's':  ignoreS = 0;  mirrorS = 1;  break;
        case 't':  ignoreT = 0;  mirrorT = 1;  break;
        case 'x':  ignoreX = 0;  mirrorX = 1;  break;
        case 'y':  ignoreY = 0;  mirrorY = 1;  break;
        case 'z':  ignoreZ = 0;  mirrorZ = 1;  break;
        case 'S':  ignoreS = 0;                break;
        case 'T':  ignoreT = 0;                break;
        case 'X':  ignoreX = 0;                break;
        case 'Y':  ignoreY = 0;                break;
        case 'Z':  ignoreZ = 0;                break;
        default:    
            fprintf(stderr,"Unsupported symmetry component '%c'\n",c);
            usage();
        }
    }
}

 void
process_commandline( int argc, char** argv ) {

    // Set up defaults:
    select_plastic();
    basisMatrix = catmull_rom;

    int  i;
    for (i = 1;   i < argc;   ++i) {
        char* a = argv[i];
        switch (*a) {

        case '-':
	    switch (*++a) {
            case '-':
	        ++a;
	        if (!strcmp( a, "debug"        )) { debug = 1;                continue; }
	        if (!strcmp( a, "checker"      )) { chequer_all_surfaces = 1; continue; }
	        if (!strcmp( a, "plastic"      )) { select_plastic();         continue; }
	        if (!strcmp( a, "constant"     )) { select_constant();        continue; }
	        if (!strcmp( a, "roughmetal"   )) { select_roughmetal();      continue; }
	        if (!strcmp( a, "shinymetal"   )) { select_shinymetal();      continue; }
	        if (!strcmp( a, "shiny"        )) { select_shiny();           continue; }
	        if (!strcmp( a, "shinyplastic" )) { select_shinyplastic();    continue; }
	        if (!strcmp( a, "clay"         )) { select_clay();            continue; }
	        if (!strcmp( a, "matte"        )) { select_matte();           continue; }
	        if (!strcmp( a, "veinedmarble" )) { select_veinedmarble();    continue; }
	        if (!strcmp( a, "greenmarble"  )) { select_greenmarble();     continue; }
	        if (!strcmp( a, "wood2"        )) { select_wood2();           continue; }
	        if (!strcmp( a, "oak"          )) { select_oak(a);            continue; }
	        if (!strcmp( a, "roughoak"     )) { select_oak(a);            continue; }
	        if (!strcmp( a, "glass"        )) { select_glass();           continue; }
	        if (!strcmp( a, "superplank"   )) { select_superplank();      continue; }
	        if (!strcmp( a, "screen"       )) { select_screen();          continue; }
	        if (!strcmp( a, "bspline"      )) { basisMatrix = bspline;    continue; }
	        if (!strcmp (a, "varyOnT"      )) { varyOnT     = 1;          continue; }
	        if (!strcmp (a, "caustics"     )) { caustics    = 1;          continue; }
                if (!strcmp (a, "nofloor"      )) { nofloor     = 1;          continue; }
                if (!strcmp (a, "nofill"       )) { nofill      = 1;          continue; }

               
	        if (!strncmp (a, "color=",6)) { select_color(a+6); continue; }
	        if (!strncmp (a, "savefile=",9)) { save_file=a+9; continue; }
	        if (!strncmp (a, "loadfile=",9)) { load_file=a+9; continue; }
	        if (!strncmp (a, "seed="  , 5 )) { surf[0].seed=surf[1].seed=
                                                   surf[2].seed=surf[3].seed=
                                                   surf[4].seed=surf[5].seed = a+5;    continue; }
	        if (!strncmp (a, "seed1=" , 6 )) { surf[0].seed = a+6; continue; }
	        if (!strncmp (a, "seed2=" , 6 )) { surf[1].seed = a+6; continue; }
	        if (!strncmp (a, "seed3=" , 6 )) { surf[2].seed = a+6; continue; }
	        if (!strncmp (a, "seed4=" , 6 )) { surf[3].seed = a+6; continue; }
	        if (!strncmp (a, "seed5=" , 6 )) { surf[4].seed = a+6; continue; }
	        if (!strncmp (a, "seed6=" , 6 )) { surf[5].seed = a+6; continue; }
	        if (!strncmp (a, "seed7=" , 6 )) { surf[6].seed = a+6; continue; }
	        if (!strncmp (a, "layers=", 7 )) { layers = atoi(a+7); assert(layers && layers <= MAX_LAYERS);   continue; }
	        if (!strncmp (a, "lumpy=",  6 )) { do_lumpy(a+6);  continue; }
	        if (!strncmp (a, "lumpy1=", 7 )) { do_lumpyn(a+7,0);  continue; }
	        if (!strncmp (a, "lumpy2=", 7 )) { do_lumpyn(a+7,1);  continue; }
	        if (!strncmp (a, "lumpy3=", 7 )) { do_lumpyn(a+7,2);  continue; }
	        if (!strncmp (a, "lumpy4=", 7 )) { do_lumpyn(a+7,3);  continue; }
	        if (!strncmp (a, "lumpy5=", 7 )) { do_lumpyn(a+7,4);  continue; }
	        if (!strncmp (a, "lumpy6=", 7 )) { do_lumpyn(a+7,5);  continue; }
	        if (!strncmp (a, "lumpy7=", 7 )) { do_lumpyn(a+7,6);  continue; }
	        if (!strncmp (a, "thick=",  6 )) { do_thickness(a+6);  continue; }
	        if (!strncmp( a, "lightradius=",12)) { do_lightradius(a+12); continue; }

	        if (!strncmp (a, "symmetry=",  9 )) { do_symmetry(a+9);  continue; }

                usage();
            }
            break;

	default: 
            if (rib_file)  usage();
            // Check for and remove ".rib" extension:
            {   int len = strlen( a );
	        if (a[len-1] != 'b'
	        ||  a[len-2] != 'i'
	        ||  a[len-3] != 'r'
	        ||  a[len-4] != '.'
                ){
                    usage();
                }
		rib_file = a;
		rib_fd = fopen( rib_file, "w" );
		if (!rib_fd) {
		    fprintf( stderr, "Couldn't create %s, aborting run\n", rib_file );
		    exit( 1 );
		}
                strcpy( tif_file, rib_file );
                tif_file[len-3]='t';
                tif_file[len-1]='f';
            }
        }
    }
    if (!rib_file)  usage();
}


 int
main( int argc, char** argv ) {
    init();
    // Must come after init() because we need
    // catmull_rom to be initialized:
    process_commandline( argc, argv );
    test();
}
